import datetime

from keg.db import db
import arrow
import pytest
import six
import sqlalchemy as sa
import sqlalchemy_utils as sautils
import keg_elements.db.mixins as mixins
import keg_elements.db.columns as columns

import kegel_app.model.entities as ents


class TestDefaultColsMixin:

    def setup_method(self, fn):
        ents.Thing.delete_cascaded()

    def test_default_odering(self):
        thing1 = ents.Thing.testing_create(id=6)
        thing2 = ents.Thing.testing_create(id=5)
        thing3 = ents.Thing.testing_create(id=7)

        assert ents.Thing.query.all() == [thing2, thing1, thing3]


class TestMethodsMixin:

    def setup_method(self, fn):
        ents.Thing.delete_cascaded()

    def test_add(self):
        ents.Thing.add(name='name', color='color', scale_check=1)
        assert ents.Thing.query.count() == 1

        row = ents.Thing.query.first()

        assert row.name == 'name'
        assert row.color == 'color'
        assert row.scale_check == 1

    def test_delete(self):
        thing = ents.Thing.testing_create()
        assert ents.Thing.query.count() == 1

        ents.Thing.delete(thing.id)
        assert ents.Thing.query.count() == 0

        assert ents.Thing.query.get(1) is None
        assert ents.Thing.delete(1) is False

    def test_delete_cascaded(self):
        ents.Thing.testing_create()
        assert ents.Thing.query.count() == 1

        ents.Thing.delete_cascaded()
        assert ents.Thing.query.count() == 0

    def test_edit(self):
        thing1 = ents.Thing.testing_create()

        ents.Thing.edit(thing1.id, name='edited')
        assert thing1.name == 'edited'

        with pytest.raises(AttributeError):
            ents.Thing.edit(name='edited')

    def test_from_dict(self):
        # Testing create uses `from_dict` so create objects by hand
        related = ents.RelatedThing(name='something')
        db.session.add(related)

        related.from_dict({
            'key_not_present': 'blah'
        })
        assert related.name == 'something'

        related.from_dict({'name': 'blah'})
        assert related.name == 'blah'

        related.from_dict({'thing': {'name': 'thing_name'}})
        assert related.thing.name == 'thing_name'

        related.thing.from_dict({'manys': [{'name': 'many_1'}, {'name': 'many_2'}]})
        assert ents.ManyToManyThing.query.count() == 2
        assert len(related.thing.manys) == 2

    def test_add_edit(self):
        obj = ents.MultiplePrimaryKeys.add_or_edit({
            'name': 'something', 'id': 1, 'other_pk': 1})

        assert obj.name == 'something'
        assert obj.id == 1
        assert obj.other_pk == 1

        with pytest.raises(sa.exc.IntegrityError):
            ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'id': 1})

        with pytest.raises(sa.exc.IntegrityError):
            ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'other_pk': 1})

        ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'id': 1, 'other_pk': 1})
        assert obj.name == 'other'

    def test_get_by(self):

        def returns_none_when_count_is_zero():
            assert ents.Thing.get_by(id=1) is None

        def returns_object_when_one_exists():
            thing = ents.Thing.testing_create()

            assert ents.Thing.get_by(id=thing.id) == thing

        def raises_when_multiple_exists():
            ents.Thing.testing_create(name='foo')
            ents.Thing.testing_create(name='foo')

            with pytest.raises(sa.orm.exc.MultipleResultsFound):
                ents.Thing.get_by(name='foo')

        returns_none_when_count_is_zero()
        returns_object_when_one_exists()
        raises_when_multiple_exists()

    def test_get_where(self):
        thing1 = ents.Thing.testing_create(name='thing1', color='black')
        ents.Thing.testing_create(name='thing2', color='black', scale_check=10)
        thing3 = ents.Thing.testing_create(name='thing3', color='orange', scale_check=10)

        assert ents.Thing.get_where(ents.Thing.name == 'thing1') == thing1
        assert ents.Thing.get_where(ents.Thing.name == 'thing5') is None
        assert (ents.Thing.get_where(ents.Thing.color == 'orange',
                                     ents.Thing.scale_check == 10) == thing3)

        with pytest.raises(sa.orm.exc.MultipleResultsFound):
            ents.Thing.get_where(ents.Thing.color == 'black')

    def test_pairs_takes_name_and_value_and_retuns_list_of_tuples(self):
        thing1 = ents.Thing.testing_create(name='A', color='orange')
        thing2 = ents.Thing.testing_create(name='B', color='black')

        expected = [(thing1.id, 'A'), (thing2.id, 'B')]
        assert ents.Thing.pairs('id', 'name') == expected

        expected = [(thing2.id, 'B'), (thing1.id, 'A')]
        assert ents.Thing.pairs('id', 'name',
                                order_by=[ents.Thing.name.desc()]) == expected

        expected = [(thing2.id, 'B')]
        query = ents.Thing.query.filter_by(name='B')
        result = ents.Thing.pairs('id', 'name', query=query)
        assert result == expected

        expected = [thing1.id, thing2.id]
        assert ents.Thing.pairs('id', 'name',
                                items=lambda obj: obj.id) == expected

    def test_to_dict(self):
        obj = ents.Thing.testing_create()

        expected = {'id', 'name', 'color', 'scale_check', 'updated_utc', 'created_utc'}
        assert set(obj.to_dict().keys()) == expected
        assert set(obj.to_dict(exclude={'id'}).keys()) == expected - {'id'}

        hybrids, exclude = (['name_and_color'],
                            ['name', 'scale_check', 'color', 'updated_utc', 'created_utc'])
        dictionary = obj.to_dict(hybrids=hybrids, exclude=exclude)

        assert dictionary == {
            'id': obj.id,
            'name_and_color': obj.name_and_color
        }

    def test_random_data_for_column(self):
        func = mixins.MethodsMixin.random_data_for_column

        assert type(func(sa.Column(sa.Unicode), 0, 0)) == six.text_type
        assert type(func(sa.Column(sa.String), 0, 0)) == six.text_type
        assert type(func(sa.Column(sa.Integer), 0, 0)) == int
        assert type(func(sa.Column(sa.Boolean), 0, 0)) == bool
        assert type(func(sa.Column(sa.Numeric), 0, 0)) == float
        assert type(func(sa.Column(sa.Date), 0, 0)) == datetime.date
        assert type(func(sa.Column(sa.DateTime), 0, 0)) == datetime.datetime
        assert type(func(sa.Column(sautils.ArrowType), 0, 0)) == arrow.Arrow
        assert isinstance(func(sa.Column(sautils.EmailType), 0, 0), six.text_type)
        assert isinstance(func(sa.Column(columns.TimeZoneType), 0, 0), str)

        with pytest.raises(ValueError):
            assert type(func(sa.Column(sa.LargeBinary), 0, 0))



"""Common SQLAlchemy column types."""

import sqlalchemy as sa


class TimeZoneType(sa.Unicode):
    """A column type for time zones, stored as a unicode string."""

    def __init__(self, length=None, **kwargs):  # Override any given length.
        super(TimeZoneType, self).__init__(length=255, **kwargs)



import kegel_app.model.entities as ents
import pytz


class TestColumns(object):

    def test_testing_create_for_timezone_columns(self):
        obj = ents.ColumnTester.testing_create()
        assert obj.time_zone in pytz.common_timezones

        obj.time_zone = 'other'
        assert obj.time_zone == 'other'
        assert obj.__mapper__.columns.time_zone.type.length == 255



import math
import wrapt
import sqlalchemy as sa
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import DateTime

from blazeutils.strings import randchars

from keg.db import db


class utcnow(expression.FunctionElement):
    type = DateTime()


@compiles(utcnow, 'postgresql')
def _pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, 'mssql')
def _ms_utcnow(element, compiler, **kw):
    return "GETUTCDATE()"


@compiles(utcnow, 'sqlite')
def _sqlite_utcnow(element, compiler, **kw):
    return "CURRENT_TIMESTAMP"


@wrapt.decorator
def no_autoflush(wrapped, instance, args, kwargs):
    autoflush = db.session.autoflush
    db.session.autoflush = False
    try:
        return wrapped(*args, **kwargs)
    finally:
        db.session.autoflush = autoflush


def validate_unique_exc(exc):
    return _validate_unique_msg(db.engine.dialect.name, str(exc))


def _validate_unique_msg(dialect, msg):
    """
        Does the heavy lifting for validate_unique_exception().

        Broken out separately for easier unit testing.  This function takes string args.
    """
    if 'IntegrityError' not in msg:
        raise ValueError('"IntegrityError" exception not found')
    if dialect == 'postgresql':
        if 'duplicate key value violates unique constraint' in msg:
            return True
    elif dialect == 'mssql':
        if 'Cannot insert duplicate key' in msg:
            return True
    elif dialect == 'sqlite':
        if 'UNIQUE constraint failed' in msg:
            return True
    else:
        raise ValueError('is_unique_exc() does not yet support dialect: %s' % dialect)
    return False


def randemail(length, randomizer=randchars):
    """Generate a random email address at the given length.
    :param length: must be at least 7 or the funuction will throw a ValueError.
    :param randomzer: is a function for generating random characters. It must have an identical
                      interface to `randchars`. The default function is `randchars`.
    """
    if length < 7:
        raise ValueError('length must be at least 7')

    half = (length - 2 - 3) / 2.0  # 2 characters for @ and . and 3 for TLD
    return (randomizer(int(math.floor(half)), 'alphanumeric')
            + '@' + randomizer(int(math.ceil(half)), 'alphanumeric')
            + '.' + randomizer(3, 'alpha'))


def session_commit():
    try:
        db.session.commit()
    except Exception:
        db.session.rollback()
        raise


def session_flush():
    try:
        db.session.flush()
    except Exception:
        db.session.rollback()
        raise


class CollectionUpdater(object):

    @no_autoflush
    def __init__(self, entity, attr_name, data):
        self.entity = entity
        self.attr_name = attr_name
        self.data = data
        self.collection = getattr(entity, attr_name)
        self.keep_children = set()

        ent_cls = entity.__class__
        queryable_attr = getattr(ent_cls, attr_name)
        # Unless they sent in the wrong type, rel_prop should be a
        # sqlalchemy.orm.properties.RelationshipProperty instance
        rel_prop = queryable_attr.property
        # now that we have the property, go through the mapper to get to the child entity class
        self.child_cls = rel_prop.mapper.class_

    @no_autoflush
    def find_child(self, data):
        """Find the child record associated with the related object"""
        child_entity_keys = self.child_cls.primary_keys()
        supplied_keys = set(data.get(column.key)
                            for column in child_entity_keys)

        if None in supplied_keys:
            return None

        for child in self.collection:
            record_keys = set(getattr(child, col.key)
                              for col in child_entity_keys)
            return child if record_keys == supplied_keys else None

    @no_autoflush
    def update(self):
        """Update the objects associated with the entity"""
        for record in self.data:
            child = self.find_child(record)
            state = sa.inspect(child) if child else None

            if state and state.persistent:
                child.edit(_commit=False, **record)
            else:
                child = self.child_cls.add(_commit=False, **record)
                self.collection.append(child)

            self.keep_children.add(child)
        self.remove_umodified()

    def remove_umodified(self):
        remove_children = [child for child in self.collection if child not in self.keep_children]
        for child in remove_children:
            self.collection.remove(child)



import wrapt


def keyword_optional(keyword, before=False, after=False, keep_keyword=False,
                     when_missing=False):
    """Execute a function before and after the decorated function if the keyword
    is in the kwargs

    Examples:
        def do_thing():
            # ... does something ...

        @keyword_optional('_do_thing', before=do_thing)
        def func(data):
            return data

        func(data, _do_thing=True)
    """

    @wrapt.decorator
    def _execute(wrapped, instance, args, kwargs):
        do_it = (kwargs.get(keyword, when_missing)
                 if keep_keyword
                 else kwargs.pop(keyword, when_missing))

        if before and do_it:
            before()

        result = wrapped(*args, **kwargs)

        if after and do_it:
            after()

        return result

    return _execute



from collections import namedtuple

from keg import current_app
import sqlalchemy as sa
from sqlalchemy_utils import ArrowType

from .db.utils import validate_unique_exc
from .db.mixins import DefaultColsMixin

ColumnCheck = namedtuple('ColumnCheck', 'name, required, fk, unique, timestamp')
ColumnCheck.__new__.__defaults__ = (True, None, None, None)


class EntityBase(object):
    entity_cls = None
    column_checks = None
    timestamp_cols = ('created_utc', 'updated_utc')
    # Set to: 'setup_class', 'setup', or None
    delete_all_on = 'setup_class'

    @classmethod
    def setup_class(cls):
        assert current_app.db_enabled, 'app.db_enabled is False'
        assert cls.entity_cls, 'The entity_cls attribute must be set.'
        assert cls.column_checks, 'column_checks should be defined for this entity'
        if cls.timestamp_cols:
            if cls.column_checks is None:
                cls.column_checks = []
            createdts_col_name, updatedts_col_name = cls.timestamp_cols
            cls.column_checks.append(ColumnCheck(createdts_col_name, timestamp='create'))
            cls.column_checks.append(ColumnCheck(updatedts_col_name, timestamp='update'))
        if cls.delete_all_on:
            cls.ent_delete_all()

    @classmethod
    def ent_delete_all(cls):
        if hasattr(cls.entity_cls, 'delete_cascaded'):
            cls.entity_cls.delete_cascaded()
        else:
            cls.entity_cls.delete_all()

    def setup(self):
        if self.delete_all_on == 'setup':
            self.ent_delete_all()

    def test_add(self):
        self.ent_delete_all()
        o = self.entity_cls.testing_create()
        assert self.entity_cls.query.count() == 1
        if hasattr(self.entity_cls, 'id'):
            assert o.id

    def check_column_null(self, col, is_required):
        assert col.nullable != is_required

    def check_column_unique(self, col, is_unique):
        assert col.unique == is_unique, 'Expected column "{}" to have unique={}'.format(
            col.name, is_unique)

    def check_column_fk(self, col, fk):
        fk_count = len(col.foreign_keys)
        if fk:
            assert fk_count == 1, 'check_column_fk() can not handle colums w/ multiple FKs'
            assert fk == list(col.foreign_keys)[0]._get_colspec()
        else:
            assert not fk_count, 'Didn\'t expect column "{}" to have a foreign key'.format(col.name)

    def check_column_timestamp(self, col, ts_type):
        if not isinstance(col.type, (ArrowType, sa.DateTime)):
            raise AssertionError('Column "{}" is not a recognized timestamp type.'.format(col.name))
        assert col.default
        if ts_type == 'update':
            assert col.onupdate, 'Column "{}" should have onupdate set'.format(col.name)
        assert col.server_default, 'Column "{}" should have server_default set'.format(col.name)

    def test_all_columns_are_constraint_tested(self):
        """Checks that all fields declared on entity are in the constraint tests"""

        expected_columns = [col.name for col in self.entity_cls.__table__.columns]
        constraint_columns = [col[0] for col in self.column_checks]
        inherited_columns = []

        # Even if the entity class is based (eventually) on DefaultColsMixin, polymorphic subclasses
        #   will not have the default cols, as the parent entity alone will have them
        if isinstance(self.entity_cls(), DefaultColsMixin) and not \
                getattr(self.entity_cls, '__mapper_args__', {}).get('polymorphic_identity', None):
            # Include columns from common import base classes
            for field in vars(DefaultColsMixin):

                # Ignore the usual __class__, __dict__, blah, blah blah
                if field.startswith('__'):
                    continue

                col = getattr(DefaultColsMixin, field)

                # Only consider SQLAlchemy columns
                if isinstance(col, sa.sql.schema.Column):
                    inherited_columns.append(field)

        combined_columns = set(constraint_columns + inherited_columns)
        if len(combined_columns) != len(expected_columns):

            missing = set(expected_columns) - combined_columns

            raise AssertionError(
                'Missing {} constraint tests for {}.'.format(
                    self.entity_cls.__name__,
                    ', '.join(missing))
            )

    def test_column_checks(self):
        if not self.column_checks:
            return
        for col_check in self.column_checks:
            col = getattr(self.entity_cls, col_check.name)
            yield self.check_column_null, col, col_check.required
            yield self.check_column_unique, col, col_check.unique
            yield self.check_column_fk, col, col_check.fk
            if col_check.timestamp:
                yield self.check_column_timestamp, col, col_check.timestamp

    def check_unique_constraint(self, **kwargs):
        self.entity_cls.testing_create(**kwargs)
        try:
            self.entity_cls.testing_create(**kwargs)
            raise AssertionError('Uniqueness error was not encountered.')
        except Exception as e:
            if not validate_unique_exc(e):
                raise



from __future__ import absolute_import

import itertools

import keg
from keg_elements.forms import Form
from pyquery import PyQuery
from wtforms import (
    BooleanField,
    RadioField,
    StringField,
)


class TemplateTest(object):
    def render(self, filename, args=None):
        template = keg.current_app.jinja_env.get_template(filename)
        return PyQuery(template.render(**(args or {})))


class TestGenericTemplates(TemplateTest):
    def test_form_field_value_macro(self):
        class TestForm(Form):
            test = StringField()

        value = 'ablewasiereisawelba'

        response = self.render('form-field-value-macro.html', {
            'form': TestForm(test=value),
            'field_name': 'test'
        })

        assert response('#dynamic #default #test')[0].value == value
        assert response('#dynamic #doubled #test')[0].value == value  # dynamic is unaffected

        assert response('#static #default #test').text() == value
        assert response('#static #doubled #test').text() == value * 2   # static is doubled

    def test_section_macro(self):
        class TestForm(Form):
            test = StringField()

        response = self.render('section-macro.html', {
            'form': TestForm(test='test-value'),
            'field_name': 'test'
        })

        assert response('#dynamic #with-caller #test')[0].value == 'test-value'
        assert response('#dynamic #with-field-name #test')[0].value == 'test-value'

        assert response('#static #with-caller #test').text() == 'test-value'
        assert response('#static #with-field-name #test').text() == 'test-value'

    def test_checkbox_with_description(self):
        class TestForm(Form):
            test = BooleanField(description='A description')

        response = self.render('generic-form.html', {
            'form': TestForm(test=True),
        })

        assert response('#dynamic .description')
        assert response('#static .description')


class TestReadonlyOrDisabledFormRender(TemplateTest):
    class MyTestForm(Form):
        radio_options = ['A', 'B', 'C']
        radio_choices = [(x, x) for x in radio_options]

        def field_flag(flags):
            """Creates a no-op validator that has the given field flags."""
            class FlaggedValidator(object):
                field_flags = flags

                def __call__(self, form, field):
                    pass

            return FlaggedValidator()

        text = StringField('Text')
        text_readonly = StringField('Read-only Text', [field_flag({'readonly'})])
        text_disabled = StringField('Disabled Text', [field_flag({'disabled'})])

        radio = RadioField('Radio', choices=radio_choices)
        radio_readonly = RadioField('Read-only Radio', [field_flag({'readonly'})],
                                    choices=radio_choices)
        radio_disabled = RadioField('Disabled Radio', [field_flag({'disabled'})],
                                    choices=radio_choices)

    def test_input_fields(self):
        def assert_string_field_attr(field_name, attr_name, attr_value):
            response = self.render('generic-input-field.html', {
                'form': self.MyTestForm(**{field_name: 'some-data'}),
                'field_name': field_name
            })

            assert response('#static #' + field_name).attr(attr_name) is None
            assert response('#dynamic #' + field_name).attr(attr_name) == attr_value

        assert_string_field_attr('text', 'readonly', None)
        assert_string_field_attr('text', 'disabled', None)

        assert_string_field_attr('text_readonly', 'readonly', 'readonly')
        assert_string_field_attr('text_readonly', 'disabled', None)

        assert_string_field_attr('text_disabled', 'readonly', None)
        assert_string_field_attr('text_disabled', 'disabled', 'disabled')

    def render_radio_field(self, field_name, field_value):
        return self.render('generic-radio-field.html', {
            'form': self.MyTestForm(**{field_name: field_value}),
            'field_name': field_name
        })

    def test_normal_radio(self):
        response = self.render_radio_field('radio', 'A')

        assert response('#dynamic input[value=A]').attr.checked == 'checked'

        for root, label in itertools.product({'#static', '#dynamic'},
                                             self.MyTestForm.radio_options):
            sel = '{} input[value={}]'.format(root, label)
            response(sel).attr.readonly is None
            response(sel).attr.disabled is None
            response(sel).attr.readonly is None
            response(sel).attr.disabled is None

    def test_readonly_radio(self):
        response = self.render_radio_field('radio_readonly', 'B')
        radio_at_value = lambda value: response('#dynamic input[value={}]'.format(value))

        assert radio_at_value('B').attr.checked == 'checked'
        assert not radio_at_value('A').attr.readonly
        assert not radio_at_value('B').attr.readonly
        assert not radio_at_value('C').attr.readonly
        assert radio_at_value('A').attr.disabled
        assert not radio_at_value('B').attr.disabled
        assert radio_at_value('C').attr.disabled

        assert response('#static p').text() == 'B'

    def test_disabled_radio(self):
        response = self.render_radio_field('radio_disabled', 'C')
        radio_at_value = lambda value: response('#dynamic input[value={}]'.format(value))

        assert radio_at_value('C').attr.checked == 'checked'
        assert not radio_at_value('A').attr.readonly
        assert not radio_at_value('B').attr.readonly
        assert not radio_at_value('C').attr.readonly
        assert radio_at_value('A').attr.disabled
        assert radio_at_value('B').attr.disabled
        assert radio_at_value('C').attr.disabled

        assert response('#static p').text() == 'C'


class TestFieldMacros(TemplateTest):
    class MyTestForm(Form):
        myfield = StringField('My Field')

    def test_div_form_group(self):
        response = self.render('field-macros.html', {
            'form': self.MyTestForm(myfield='My Data'),
            'field_name': 'myfield',
            'form_group_class': None,
            'field_kwargs': {},
        })

        assert response('#dynamic #div_form_group div.form-group').text() == 'Contents'
        assert response('#static  #div_form_group div.form-group').text() == 'Contents'

        assert response('#dynamic #field_widget #myfield')[0].value == 'My Data'
        assert response('#static  #field_widget #myfield').text() == 'My Data'



import datetime as dt
import operator
import random

from keg.db import db
from sqlalchemy_utils import ArrowType, EmailType
import arrow
import blazeutils.strings
import pytz
import six
import sqlalchemy as sa

import keg_elements.decorators as decor
import keg_elements.db.columns as columns
import keg_elements.db.utils as dbutils


might_commit = decor.keyword_optional('_commit', after=dbutils.session_commit, when_missing=True)
might_flush = decor.keyword_optional('_flush', after=dbutils.session_flush)


class DefaultColsMixin(object):
    id = sa.Column('id', sa.Integer, primary_key=True)
    created_utc = sa.Column(ArrowType, nullable=False, default=arrow.now,
                            server_default=dbutils.utcnow())
    updated_utc = sa.Column(ArrowType, nullable=False, default=arrow.now, onupdate=arrow.now,
                            server_default=dbutils.utcnow())


class MethodsMixin:
    def from_dict(self, data):
        """Update the instance with the passed information in `data`.

        `from_dict` will also update relationships either, 1:1, 1:N, M:M.

        .. note::
            `from_dict` intentionally does not commit and for related entities
            turns off auto flushing. This is to prevent premature flushes with
            incomplete objects
        """

        @dbutils.no_autoflush
        def update_related(key, value, prop):
            """Update the entity based on the type of relationship it is"""
            related_class = prop.mapper.class_

            if prop.uselist:
                new_list = list()
                for row in value:
                    obj = related_class.add_or_edit(row, _commit=False)
                    new_list.append(obj)
                setattr(self, key, new_list)
            else:
                setattr(self, key, related_class.add_or_edit(value, _commit=False))

        mapper = self.__mapper__
        entity_props = {attr.key: attr for attr in mapper.attrs}

        for key, value in six.iteritems(data):
            prop = entity_props.get(key)
            is_column = isinstance(prop, sa.orm.properties.ColumnProperty)
            is_relationship = isinstance(prop, sa.orm.properties.RelationshipProperty)

            if prop is None:
                continue
            elif is_column:
                setattr(self, key, value)
            elif is_relationship:
                update_related(key, value, prop)
            else:
                raise NotImplementedError(
                    'Updating {} property types is not implemented.')

    def to_dict(self, exclude=frozenset(), hybrids=frozenset()):
        """Covert the object properties to a dictionary.

        :param exclude: a list of columns to ignore
        :param hybrids: a list of the hybrid properties to include in the dictionary.
        :returns: a dictionary representation of the object

        .. note: By default hybrid properties are not included in the returned dict. To add a hybrid
        property to the returned dict pass a list of the property names and they will be included.
        """
        data = dict((name, getattr(self, name))
                    for name in self.column_names()
                    if name not in exclude)

        for hybrid in hybrids:
            data[hybrid] = getattr(self, hybrid)

        return data

    @classmethod
    def column_names(cls):
        return {col.key for col in cls.__mapper__.columns}

    @classmethod
    def primary_keys(cls):
        return cls.__table__.primary_key.columns

    @might_commit
    @might_flush
    @classmethod
    def add(cls, **kwargs):
        obj = cls()
        obj.from_dict(kwargs)
        db.session.add(obj)
        return obj

    @might_commit
    @might_flush
    @classmethod
    def delete(cls, oid):
        """Delete an object from the session

        :param oid: the object identifier, normally the primary key
        :rtype: bool
        :return: The result of the operation
        """
        obj = cls.query.get(oid)

        if obj is None:
            return False

        db.session.delete(obj)
        return True

    @might_commit
    @classmethod
    def delete_cascaded(cls):
        cls.query.delete(synchronize_session=False)
        db.session.expire_all()

    @might_commit
    @might_flush
    @classmethod
    def edit(cls, oid=None, **kwargs):
        try:
            primary_keys = oid or [kwargs.get(x.name)
                                   for x in cls.primary_keys()
                                   if x is not None]
        except KeyError:
            raise AttributeError('No primary key was found in `oid` or `kwargs`'
                                 ' for which to retrieve the object to edit')

        obj = cls.query.get(primary_keys)
        obj.from_dict(kwargs)
        return obj

    @classmethod
    def get_by(cls, **kwargs):
        """Returns the instance of this class matching the given criteria or
        None if there is no record matching the criteria.

        If multiple records are returned, an exception is raised.
        """
        return cls.query.filter_by(**kwargs).one_or_none()

    @classmethod
    def get_where(cls, *clauses):
        """
        Returns the instance of this class matching the given clause(s) or None
        if there is no record matching the criteria.

        If multiple records are returned, an exception is raised.
        """
        return cls.query.filter(*clauses).one_or_none()

    @classmethod
    def pairs(cls, key_field, value_field, order_by=(), query=None,
              items=None):
        """Return a list of two item tuples

        :param key_field: string representing the key
        :param value_field: string representing the value
        :param order_by: iterable of columns to order the query by
        :param query: a base query from which to generate the pairs
        :param items: a function which takes one record returned by query and
                      returns the tuple object
        """

        items = items if items else operator.attrgetter(key_field, value_field)
        query = query or cls.query
        result = query.order_by(*order_by).all()

        return [items(obj) for obj in result]

    @classmethod
    def testing_create(cls, **kwargs):
        """Create an object for testing with default data appropriate for the field type

        * Will automatically set most field types ignoring those passed in via kwargs.
        * Subclasses that have foreign key relationships should setup those relationships before
          calling this method.

        Special kwargs:
        _numeric_defaults_range: a tuple of (HIGH, LOW) which controls the acceptable defaults of
                                 the two number types. Both integer and numeric (float) fields are
                                 controlled by this setting.
        """

        NUMERIC_HIGH, NUMERIC_LOW = kwargs.pop('_numeric_defaults_range', (-100, 100))

        insp = sa.inspection.inspect(cls)

        skippable = lambda column: (column.key in kwargs      # skip fields already in kwargs
                                    or column.foreign_keys    # skip foreign keys
                                    or column.server_default  # skip fields with server defaults
                                    or column.default         # skip fields with defaults
                                    or column.primary_key     # skip any primary key
                                    )

        for column in (col for col in insp.columns if not skippable(col)):
            try:
                kwargs[column.key] = cls.random_data_for_column(
                    column, NUMERIC_HIGH, NUMERIC_LOW)
            except ValueError:
                pass

        return cls.add(**kwargs)

    @classmethod
    def random_data_for_column(cls, column, numeric_high, numeric_low):
        if isinstance(column.type, sa.types.Enum):
            return random.choice(column.type.enums)
        elif isinstance(column.type, sa.types.Boolean):
            return random.choice([True, False])
        elif isinstance(column.type, sa.types.Integer):
            return random.randint(numeric_high, numeric_low)
        elif isinstance(column.type, sa.types.Numeric):
            return random.uniform(numeric_high, numeric_low)
        elif isinstance(column.type, sa.types.Date):
            return dt.date.today()
        elif isinstance(column.type, sa.types.DateTime):
            return dt.datetime.utcnow()
        elif isinstance(column.type, ArrowType):
            return arrow.utcnow()
        elif isinstance(column.type, EmailType):
            return dbutils.randemail(min(column.type.length or 50, 50))
        elif isinstance(column.type, columns.TimeZoneType):
            return random.choice(pytz.common_timezones)
        elif isinstance(column.type, (sa.types.String, sa.types.Unicode)):
            return blazeutils.strings.randchars(min(column.type.length or 25, 25))
        raise ValueError('No randomization for this column type')

    @might_commit
    @might_flush
    @classmethod
    def add_or_edit(cls, data):
        """Creates or updates the record associated with `data`

        `add_or_edit` supports multiple primary key entities.
        """
        if isinstance(data, db.Model) or not data:
            return data

        primary_keys = [data.get(x.name) for x in cls.primary_keys() if x
                        is not None]
        obj = cls.query.get(primary_keys)

        return (cls.add(_commit=False, **data)
                if obj is None
                else cls.edit(primary_keys, _commit=False, **data))

    def update_collection(self, attr_name, data):
        dbutils.CollectionUpdater(self, attr_name, data).update()


class DefaultMixin(DefaultColsMixin, MethodsMixin):
    pass



import pytest
import validators
import keg

from keg_elements.db.utils import randemail
import kegel_app.model.entities as ents


def test_randemail():
    for length in range(0, 6):
        with pytest.raises(ValueError):
            randemail(length)

    def not_so_rand_email(length, char):
        return randemail(length, randomizer=lambda n, *args, **kwargs: char * n)

    assert not_so_rand_email(7, 'a') == 'a@a.aaa'
    assert not_so_rand_email(8, 'b') == 'b@bb.bbb'
    assert not_so_rand_email(9, 'c') == 'cc@cc.ccc'

    # Fuzzy testing for an extra dose of confidence
    for length in range(7, 50):
        email = randemail(length)
        assert len(email) == length
        assert validators.email(email)

    # Get some confidence of randomness
    assert len(set(randemail(50) for _ in range(5))) == 5, \
        'randemail not random (beware non-determinism; try again)'


class TestUpdateCollection(object):

    def setup_method(self, method):
        keg.db.db.session.remove()
        ents.Thing.delete_cascaded()
        ents.RelatedThing.delete_cascaded()

    def test_add(self):
        thing = ents.Thing.testing_create()
        assert not thing.related_things

        data = [{'name': 'test', 'is_enabled': True}]
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 1
        related = thing.related_things[0]
        assert related.name == 'test'
        assert related.is_enabled

    def test_edit(self):
        thing = ents.Thing.testing_create()
        related = ents.RelatedThing.testing_create(thing=thing, name='a')

        data = [
            {'id': related.id, 'name': 'b', 'is_enabled': False}
        ]
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 1
        related = thing.related_things[0]
        assert related.name == 'b'
        assert not related.is_enabled

    def test_append_id_none(self):
        thing = ents.Thing.testing_create()
        related = ents.RelatedThing.testing_create(thing=thing, name='a',
                                                   is_enabled=True)

        data = [
            {'is_enabled': False}
        ]
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 1
        related = thing.related_things[0]

        assert related.name is None
        assert not related.is_enabled

    def test_remove(self):
        thing = ents.Thing.testing_create()
        ents.RelatedThing.testing_create(thing=thing, name='a', is_enabled=True)
        assert len(thing.related_things) == 1

        data = []
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 0



import os.path as osp
from setuptools import setup, find_packages

cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'readme.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

setup(
    name="KegElements",
    description=("A testing ground for Keg related code and ideas."),
    long_description='\n\n'.join((README, CHANGELOG)),
    author="Randy Syring",
    author_email="randy.syring@level12.io",
    url='https://github.com/level12/keg-elements',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    license='BSD',
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    setup_requires=[
        'setuptools_scm',
    ],
    use_scm_version=True,
    install_requires=[
        'arrow',
        'Flask-WTF',
        'Keg',
        'pytz',
        'WTForms-Alchemy',
    ]
)



import pytest
import wtforms
import keg_elements.forms.validators as validators


class Form():
    def __init__(self, obj):
        self.obj = obj

    def get_object_by_field(self, field):
        return 1


class TestUniqueValidator(object):
    def test_unique_validator(self):
        field = wtforms.fields.StringField('thing')
        validator = validators.ValidateUnique(object_html_link=lambda field: 'link')

        assert validator(Form(1), field) is True

        with pytest.raises(wtforms.ValidationError,
                           message=('This valuue must be unique bit is already '
                                    'assigned link.')):
            assert validator(Form(2), field)



class TestProfile(object):
    KEG_KEYRING_ENABLE = False

    # Turn off CSRF protection for tests.  Makes form testing easier and provides no value during
    # testing.
    WTF_CSRF_ENABLED = False
    SQLALCHEMY_TRACK_MODIFICATIONS = True



from __future__ import absolute_import
from __future__ import unicode_literals

import functools
import inspect
import logging

import flask
from flask_wtf import Form as BaseForm
from keg.db import db
import sqlalchemy as sa
import six
import wtforms.fields
import wtforms.form
from wtforms.validators import InputRequired, Optional
from wtforms_alchemy import model_form_factory, FormGenerator as FormGeneratorBase
from wtforms_components.fields import SelectField as SelectFieldBase

from keg_elements.forms.validators import NumberScale

form_element = flask.Blueprint('form_element', __name__)
log = logging.getLogger(__name__)


def to_title_case(x):
    """ underscore or dash to title case notation """
    return x.replace('_', ' ').replace('-', ' ').title()


# sentinel
_not_given = ()


class FieldMeta(object):

    def __init__(self, label_text=_not_given, description=_not_given, label_modifier=_not_given,
                 choices_modifier=_not_given, choices=None, required=_not_given, widget=_not_given,
                 extra_validators=[]):
        self.label_text = label_text
        self.label_modifier = label_modifier
        self.description = description
        self.choices_modifier = choices_modifier
        self.choices = choices
        self.required = required
        self.widget = widget
        self.extra_validators = extra_validators

        assert self.required in (_not_given, False, True)

    def apply_to_field(self, field):
        # field is a wtforms.fields.core.UnboundField instance
        self.apply_to_label(field)
        self.apply_to_description(field)
        self.apply_to_choices(field)
        self.apply_required(field)
        self.apply_widget(field)
        self.apply_extra_validators(field)

    def apply_to_label(self, field):
        default_label = field.kwargs['label']
        if self.label_text is not _not_given:
            label_text = self.label_text
        elif self.label_modifier is None:
            label_text = default_label
        elif self.label_modifier is _not_given:
            label_text = to_title_case(default_label)
        else:
            label_text = self.label_modifier(default_label)
        field.kwargs['label'] = self.modify_label(label_text)

    def modify_label(self, label_text):
        """ for subclasses to easily modify the final label text value """
        return label_text

    def apply_to_description(self, field):
        default_description = field.kwargs['description']
        if self.description is _not_given:
            description = default_description
        else:
            description = self.description
        field.kwargs['description'] = self.modify_description(description)

    def modify_description(self, description):
        """ for subclasses to easily modify the final label text value """
        return description

    def apply_to_choices(self, field):
        default_choices = field.kwargs.get('choices', None)
        if default_choices is None:
            # this isn't a field that has choices
            return

        if self.choices_modifier is None:
            modifier = None
        elif self.choices_modifier is not _not_given:
            modifier = self.choices_modifier
        elif self.label_modifier is None:
            # no choices modifier and the label modifier is explicit, so no label modifier
            modifier = None
        elif self.label_modifier is _not_given:
            # title case to labels by default
            modifier = to_title_case
        else:
            # a label modifier was given, use that since no choices modifier was given to override
            modifier = self.label_modifier

        if self.choices is not None:
            choices = self.choices
        elif modifier is None:
            choices = default_choices
        else:
            map_func = lambda pair: (pair[0], modifier(pair[1]))
            choices = map(map_func, default_choices)

        field.kwargs['choices'] = self.modify_choices(choices)

    def modify_choices(self, choices):
        return choices

    def apply_required(self, field):
        validators = field.kwargs.get('validators', [])

        if self.required == _not_given:
            # required value not given on FieldMeta, don't make any changes
            pass
        elif self.required:
            # If a required validator isn't present, we need to add one.
            req_val_test = lambda val: hasattr(val, 'field_flags') and 'required' in val.field_flags
            if not list(filter(req_val_test, validators)):
                validators.append(InputRequired())

            # If an optional validator is present, we need to remove it.
            not_opt_val_test = lambda val: not hasattr(val, 'field_flags') or \
                'optional' not in val.field_flags
            not_opt_validators = list(filter(not_opt_val_test, validators))
            field.kwargs['validators'] = not_opt_validators
        else:
            # If an optional validator isn't present, we need to add one.
            opt_val_test = lambda val: hasattr(val, 'field_flags') and 'optional' in val.field_flags
            if not list(filter(opt_val_test, validators)):
                validators.append(Optional())

            # If a required validator is present, we need to remove it.
            non_req_val_test = lambda val: not hasattr(val, 'field_flags') or \
                'required' not in val.field_flags
            not_req_validators = list(filter(non_req_val_test, validators))
            field.kwargs['validators'] = not_req_validators

    def apply_widget(self, field):
        if self.widget != _not_given:
            field.kwargs['widget'] = self.widget

    def apply_extra_validators(self, field):
        field.kwargs.setdefault('validators', [])
        field.kwargs['validators'] += self.extra_validators


def select_coerce(es_pass_thru, coerce, value):
    if es_pass_thru and value == '':
        return value
    if coerce is not _not_given:
        return coerce(value)

    # try coercing to int first.  If not valid, fall back to default behavior
    try:
        return int(value)
    except ValueError as e:
        if 'invalid literal for int()' not in str(e):
            raise
        return six.text_type(value)


class SelectField(SelectFieldBase):
    """
        Provides helpful features above wtforms_components SelectField which it is based on:

        1) Adds a blank choice by default at the front of the choices.  This results in your user
           being forced to select something if the field is required, which avoids unintial
           deaulting of the first value in the field getting submitted.
        2) The coerce function used for the choices will automatically convert to int if possible,
           falling back to unicode if the value is not an integer.
    """
    def __init__(self, *args, **kwargs):
        self.add_blank_choice = kwargs.pop('add_blank_choice', True)
        coerce_arg = kwargs.pop('coerce', _not_given)
        super(SelectField, self).__init__(*args, **kwargs)

        if self.add_blank_choice:
            # If we are adding a blank choice, and it is selected, we want the value that comes back
            # in .data to be None -> as if no value was selected.
            #
            # self.filters is a tuple, so have to do some extra work.
            self.filters = [lambda x: None if x == '' else x] + list(self.filters)

        self.coerce = functools.partial(select_coerce, self.add_blank_choice, coerce_arg)

    def iter_choices(self):
        if self.add_blank_choice:
            yield ('', '', (self.coerce, False))
        for value in super(SelectField, self).iter_choices():
            yield value

    @property
    def choice_values(self):
        values = super(SelectField, self).choice_values
        if self.add_blank_choice:
            return [''] + values
        return values

    @property
    def selected_choice_label(self):
        value_dict = dict(self.concrete_choices)
        return value_dict.get(self.data)


class RequiredBoolRadioField(wtforms.fields.RadioField):
    def __init__(self, *args, **kwargs):
        true_label = kwargs.pop('true_label', 'Yes')
        false_label = kwargs.pop('false_label', 'No')

        def bool_coerce(val):
            if val == u'True':
                return True
            if val == u'False':
                return False
            return val

        kwargs['choices'] = [(True, true_label), (False, false_label)]
        kwargs['coerce'] = bool_coerce
        kwargs['validators'] = [InputRequired()] + kwargs.get('validators', [])

        super(RequiredBoolRadioField, self).__init__(*args, **kwargs)
        self.type = 'RadioField'


class FormGenerator(FormGeneratorBase):
    def __init__(self, form_class):
        super(FormGenerator, self).__init__(form_class)
        self.fields_meta = getattr(self.form_class, 'FieldsMeta', None)

    def get_field_class(self, column):
        field_cls = super(FormGenerator, self).get_field_class(column)
        if field_cls is SelectFieldBase:
            return SelectField

        is_required_boolean = (field_cls is wtforms.fields.BooleanField
                               and not column.nullable
                               and not column.default)

        if is_required_boolean:
            return RequiredBoolRadioField

        return field_cls

    def get_field_modifier(self, prop):

        # is there an entry in FieldMeta?
        if hasattr(self.fields_meta, prop.key):
            field_modifier = getattr(self.fields_meta, prop.key)
        else:
            field_modifier = getattr(self.fields_meta, '__default__', _not_given)
            if field_modifier is _not_given:
                return None

        return field_modifier() if inspect.isclass(field_modifier) else field_modifier

    def create_field(self, prop, column):
        field = super(FormGenerator, self).create_field(prop, column)
        modifier = self.get_field_modifier(prop)
        if modifier is not None:
            modifier.apply_to_field(field)
        return field

    def create_validators(self, prop, column):
        validators = super(FormGenerator, self).create_validators(prop, column)
        if isinstance(column.type, sa.Numeric):
            validators.append(NumberScale(column.type.scale))
        return validators


def field_to_dict(field):
    if isinstance(field, wtforms.fields.FormField):
        return form_fields_to_dict(field)
    if isinstance(field, wtforms.fields.FieldList):
        return [field_to_dict(subfield) for subfield in field]
    return dict(data=field.data, errors=field.errors)


def form_fields_to_dict(form):
    return dict((name, field_to_dict(field)) for name, field in six.iteritems(form._fields))


class Form(BaseForm):
    def __init__(self, *args, **kwargs):
        super(Form, self).__init__(*args, **kwargs)
        self.after_init(args, kwargs)

    def after_init(self, args, kwargs):
        pass

    def fields_todict(self):
        """Turns a form into dicts and lists with both data and errors for each field."""
        return form_fields_to_dict(self)


BaseModelForm = model_form_factory(Form, form_generator=FormGenerator)


class ModelForm(BaseModelForm):
    @classmethod
    def get_session(self):
        return db.session
